// complex-bench.c (copyleft) 2016 Claude Heiland-Allen <claude@mathr.co.uk>
// see: <https://mathr.co.uk/blog/2016-04-10_complex_squaring.html>
// gcc -std=c99 -Wall -Wextra -pedantic -O3 -march=native -o complex-bench complex-bench.c -lmpfr

#include <stdio.h>
#include <time.h>
#include <mpfr.h>

/*
1M 2S 3A 8V
x := x^2 - y^2 + cx
y := 2 x y + cy
*/
struct naive {
  mpfr_t cx, cy, x, y, x2, y2, xy, x2my2;
};

void naive_init(struct naive *s, const mpfr_t cx, const mpfr_t cy) {
  mpfr_prec_t p = mpfr_get_prec(cx);
  mpfr_init2(s->cx, p);
  mpfr_init2(s->cy, p);
  mpfr_init2(s->x, p);
  mpfr_init2(s->y, p);
  mpfr_init2(s->x2, p);
  mpfr_init2(s->y2, p);
  mpfr_init2(s->xy, p);
  mpfr_init2(s->x2my2, p);
  mpfr_set(s->cx, cx, MPFR_RNDN);
  mpfr_set(s->cy, cy, MPFR_RNDN);
  mpfr_set_si(s->x, 0, MPFR_RNDN);
  mpfr_set_si(s->y, 0, MPFR_RNDN);
}

void naive_clear(struct naive *s) {
  mpfr_clear(s->cx);
  mpfr_clear(s->cy);
  mpfr_clear(s->x);
  mpfr_clear(s->y);
  mpfr_clear(s->x2);
  mpfr_clear(s->y2);
  mpfr_clear(s->xy);
  mpfr_clear(s->x2my2);
}

void naive_step(struct naive *s) {
  mpfr_mul(s->xy, s->x, s->y, MPFR_RNDN);
  mpfr_sqr(s->x2, s->x, MPFR_RNDN);
  mpfr_sqr(s->y2, s->y, MPFR_RNDN);
  mpfr_sub(s->x2my2, s->x2, s->y2, MPFR_RNDN);
  mpfr_mul_2si(s->xy, s->xy, 1, MPFR_RNDN);
  mpfr_add(s->x, s->x2my2, s->cx, MPFR_RNDN);
  mpfr_add(s->y, s->xy, s->cy, MPFR_RNDN);
}

/*
2M 0S 4A 8V
x := (x + y) (x - y) + cx
y := 2 x y + cy
*/
struct karatsuba {
  mpfr_t cx, cy, x, y, xpy, xmy, xy, xmyxpy;
};

void karatsuba_init(struct karatsuba *s, const mpfr_t cx, const mpfr_t cy) {
  mpfr_prec_t p = mpfr_get_prec(cx);
  mpfr_init2(s->cx, p);
  mpfr_init2(s->cy, p);
  mpfr_init2(s->x, p);
  mpfr_init2(s->y, p);
  mpfr_init2(s->xpy, p);
  mpfr_init2(s->xmy, p);
  mpfr_init2(s->xy, p);
  mpfr_init2(s->xmyxpy, p);
  mpfr_set(s->cx, cx, MPFR_RNDN);
  mpfr_set(s->cy, cy, MPFR_RNDN);
  mpfr_set_si(s->x, 0, MPFR_RNDN);
  mpfr_set_si(s->y, 0, MPFR_RNDN);
}

void karatsuba_clear(struct karatsuba *s) {
  mpfr_clear(s->cx);
  mpfr_clear(s->cy);
  mpfr_clear(s->x);
  mpfr_clear(s->y);
  mpfr_clear(s->xpy);
  mpfr_clear(s->xmy);
  mpfr_clear(s->xy);
  mpfr_clear(s->xmyxpy);
}

void karatsuba_step(struct karatsuba *s) {
  mpfr_add(s->xpy, s->x, s->y, MPFR_RNDN);
  mpfr_sub(s->xmy, s->x, s->y, MPFR_RNDN);
  mpfr_mul(s->xmyxpy, s->xmy, s->xpy, MPFR_RNDN);
  mpfr_mul(s->xy, s->x, s->y, MPFR_RNDN);
  mpfr_mul_2si(s->xy, s->xy, 1, MPFR_RNDN);
  mpfr_add(s->x, s->xmyxpy, s->cx, MPFR_RNDN);
  mpfr_add(s->y, s->xy, s->cy, MPFR_RNDN);
}

/*
0M 3S 6A 9V
x := x^2 - y^2
y := (x + y)^2 - (x^2 + y^2)
*/
struct kalles {
  mpfr_t cx, cy, x, y, a, b, xpy, c, d;
};

void kalles_init(struct kalles *s, const mpfr_t cx, const mpfr_t cy) {
  mpfr_prec_t p = mpfr_get_prec(cx);
  mpfr_init2(s->cx, p);
  mpfr_init2(s->cy, p);
  mpfr_init2(s->x, p);
  mpfr_init2(s->y, p);
  mpfr_init2(s->a, p);
  mpfr_init2(s->b, p);
  mpfr_init2(s->xpy, p);
  mpfr_init2(s->c, p);
  mpfr_init2(s->d, p);
  mpfr_set(s->cx, cx, MPFR_RNDN);
  mpfr_set(s->cy, cy, MPFR_RNDN);
  mpfr_set_si(s->x, 0, MPFR_RNDN);
  mpfr_set_si(s->y, 0, MPFR_RNDN);
}

void kalles_clear(struct kalles *s) {
  mpfr_clear(s->cx);
  mpfr_clear(s->cy);
  mpfr_clear(s->x);
  mpfr_clear(s->y);
  mpfr_clear(s->a);
  mpfr_clear(s->b);
  mpfr_clear(s->xpy);
  mpfr_clear(s->c);
  mpfr_clear(s->d);
}

void kalles_step(struct kalles *s) {
  mpfr_add(s->xpy, s->x, s->y, MPFR_RNDN);
  mpfr_sqr(s->a, s->x, MPFR_RNDN);
  mpfr_sqr(s->b, s->y, MPFR_RNDN);
  mpfr_sqr(s->c, s->xpy, MPFR_RNDN);
  mpfr_sub(s->d, s->a, s->b, MPFR_RNDN);
  mpfr_add(s->x, s->d, s->cx, MPFR_RNDN);
  mpfr_add(s->d, s->a, s->b, MPFR_RNDN);
  mpfr_sub(s->a, s->c, s->d, MPFR_RNDN);
  mpfr_add(s->y, s->a, s->cy, MPFR_RNDN);
}

/*
types
*/
typedef void init_t(void *, const mpfr_t, const mpfr_t);
typedef void clear_t(void *);
typedef void step_t(void *);

/*
benchmark
*/
int main(void) {
  struct naive snaive;
  struct karatsuba skaratsuba;
  struct kalles skalles;
  void *structs[3] = { &snaive, &skaratsuba, &skalles };
  init_t *inits[3] = { naive_init, karatsuba_init, kalles_init };
  clear_t *clears[3] = { naive_clear, karatsuba_clear, kalles_clear };
  step_t *steps[3] = { naive_step, karatsuba_step, kalles_step };
  mpfr_prec_t precs[9] = { 100, 316, 1000, 3162, 10000, 31623, 100000, 316228, 1000000 };
  printf("<table><tr><td>&nbsp;</td><th>naive<br />1M 2S 3A 8V</th><th>Karatsuba<br />2M 0S 4A 8V</th><th>Kalles<br />0M 3S 6A 9V</th></tr>\n");
  for (int p = 0; p < 9; ++p) {
    printf("<tr><th>%d</th>\n", precs[p]);
    mpfr_t cx, cy;
    mpfr_init2(cx, precs[p]);
    mpfr_init2(cy, precs[p]);
    mpfr_set_si(cx, -2, MPFR_RNDN);
    mpfr_set_si(cy, -3, MPFR_RNDN);
    mpfr_exp(cx, cx, MPFR_RNDN);
    mpfr_exp(cy, cy, MPFR_RNDN);
    for (int a = 0; a < 3; ++a) {
      inits[a](structs[a], cx, cy);
      time_t start = time(0);
      time_t now = time(0);
      int count = 0;
      while (now - start < 10) {
        steps[a](structs[a]);
        count++;
        now = time(0);
      }
      printf("<td>%d</td>\n", count);
      fflush(stdout);
      clears[a](structs[a]);
    }
    mpfr_clear(cx);
    mpfr_clear(cy);
    printf("</tr>\n");
  }
  printf("</table>\n");
  return 0;
}
